"""Test functions for i_mongodb.py.
"""
from decimal import Decimal

import pytest

import i_mongodb as imdb

# initialize module variables
DB_NAME = '_testdb'


@pytest.fixture(name='mdb')
def fixture_mongodb_interface():
    """Pytest fixture to initialize and return the MongoDBInterface object.
    """
    return imdb.MongoDBInterface(db_name=DB_NAME)

def test_encode_decimal(mdb):
    """Tests inserting a document with Decimal values.
    """
    doc_write = {
        '_id': 'test_decimal',
        'decimal_value': Decimal('123.456')
    }

    doc_read = mdb._test.find_one_and_replace(
        filter={'_id': 'test_decimal'},
        replacement=doc_write,
        upsert=True)

    assert type(doc_read['decimal_value']) is Decimal


def test_decode_decimal(mdb):
    """Tests retrieving a document back into Decimal values.
    """
    doc_read = mdb._test.find_one(
        filter={'_id': 'test_decimal'}
    )

    assert doc_read
    assert type(doc_read['decimal_value']) is Decimal



"""Test functions for i_mongodb.py.
"""
from datetime import datetime
from dateutil import tz, utils

import pytest

import i_mongodb as imdb

# initialize module variables
DB_NAME = '_testdb'
DT_NAIVE = datetime(2020, 8, 24, 11, 23)
DT_UTC = utils.default_tzinfo(DT_NAIVE, tz.UTC)
DT_LOCAL = utils.default_tzinfo(DT_NAIVE, tz.tzlocal())


@pytest.fixture(name='mongodb_if')
def fixture_mongodb_interface():
    """Pytest fixture to initialize and return the MongoDBInterface object.
    """
    return imdb.MongoDBInterface(db_name=DB_NAME)

def test_init_mongodb(mongodb_if):
    """Tests MongoDB initialization.
    """
    mdb = mongodb_if.mdb
    assert mdb

    # verify database name
    assert mdb.name == DB_NAME

def test_create_collection(mongodb_if):
    """Tests collection creation.
    """
    collection_name = '_test_create'

    # delete collection if already present
    collection_name_list = mongodb_if.mdb.list_collection_names()
    if collection_name in collection_name_list:
        mongodb_if.mdb[collection_name].drop()

    # verify that the collection if not present
    collection_name_list = mongodb_if.mdb.list_collection_names()
    assert collection_name not in collection_name_list

    # create collection
    collection = mongodb_if.create_collection(collection_name)
    assert collection.name == collection_name

    # verify that the collection was created
    collection_name_list = mongodb_if.mdb.list_collection_names()
    assert collection_name in collection_name_list

def test_read_collection(mongodb_if):
    """Tests collection read.
    """
    collection_name = '_test'
    collection = mongodb_if.read_collection(collection_name)
    assert collection.name == collection_name

def test_read_collection_by_attr(mongodb_if):
    """Tests collection read.
    """
    collection_name = '_test'
    collection = mongodb_if._test
    assert collection.name == collection_name

@pytest.fixture(name='test_collection')
def fixture_datetime_test_collection(mongodb_if):
    """Pytest fixture to set a document with different datetime formats.

    Takes the mongodb_if fixture and returns a collection
    """
    doc_write = {
        '_id': 'test_datetime',
        'datetime_naive': DT_NAIVE,
        'datetime_utc': DT_UTC,
        'datetime_local': DT_LOCAL
    }

    collection = mongodb_if.mdb['_test']
    collection.find_one_and_replace(
        filter={'_id': 'test_datetime'},
        replacement=doc_write,
        upsert=True)

    return collection

def test_read_datetime_tz_utc(test_collection):
    """Tests reading UTC datetime from MongoDB.

    UTC datetime objects stored in MongoDB are retrieved naive.
    """
    doc_read = test_collection.find_one({'_id': 'test_datetime'})
    dt_read = doc_read['datetime_utc']

    assert dt_read == DT_UTC

def test_read_datetime_tz_local(test_collection):
    """Tests reading local datetime from MongoDB.

    UTC datetime objects stored in MongoDB are retrieved naive.
    """
    doc_read = test_collection.find_one({'_id': 'test_datetime'})
    dt_read = doc_read['datetime_local']

    assert dt_read == DT_LOCAL

def test_read_datetime_tz_naive(test_collection):
    """Tests reading timezone-naive datetime from MongoDB.

    UTC datetime objects stored in MongoDB are retrieved naive.
    """
    doc_read = test_collection.find_one({'_id': 'test_datetime'})
    dt_read = doc_read['datetime_naive']

    assert dt_read == DT_UTC

def test_delete_collection(mongodb_if):
    """Tests collection deletion.
    """
    collection_name = '_test_delete'

    # create collection if not already present
    mongodb_if.create_collection(collection_name)

    # verify that the collection if present
    collection_name_list = mongodb_if.mdb.list_collection_names()
    assert collection_name in collection_name_list

    # delete the collection
    mongodb_if.delete_collection(collection_name)

    # verify that the collection was deleted
    collection_name_list = mongodb_if.mdb.list_collection_names()
    assert collection_name not in collection_name_list



"""Class module to interface with MongoDB.
"""
from decimal import Decimal
import os

from aracnid_logger import Logger
from bson.codec_options import CodecOptions, TypeCodec, TypeRegistry
from bson.decimal128 import Decimal128
from dateutil import tz
import pymongo

# initialize logging
logger = Logger(__name__).get_logger()

class DecimalCodec(TypeCodec):
    python_type = Decimal    # the Python type acted upon by this type codec
    bson_type = Decimal128   # the BSON type acted upon by this type codec
    def transform_python(self, value):
        """Function that transforms a custom type value into a type that BSON can encode.
        """
        return Decimal128(value)

    def transform_bson(self, value):
        """Function that transforms a vanilla BSON type value into our
        custom type."""
        return value.to_decimal()

class MongoDBInterface:
    """MongoDB interface class.

    Environment Variables:
        MONGODB_USER_TOKEN: MongoDB username and password.
        MONGODB_HOSTNAME: MongoDB host where database is running.
        MONGODB_DBNAME: Database name.

    Attributes:
        mdb: MongoDB database
        db_name: Name of the interfacing database.
        mongo_client: MongoDB client.

    Exceptions:
        DuplicateKeyError: MongoDB duplicate key error
    """

    DuplicateKeyError = pymongo.errors.DuplicateKeyError


    def __init__(self, db_name=None):
        """Initializes the interface with the database name.

        If no database name is supplied, the name is read from environment.

        Args:
            db_name: The name of the interfacing database.
        """
        # read environment variables
        mdb_user_token = os.environ.get('MONGODB_USER_TOKEN')
        mdb_hostname = os.environ.get('MONGODB_HOSTNAME')
        self.db_name = os.environ.get('MONGODB_DBNAME')

        # override database name, if provided
        if db_name:
            self.db_name = db_name

        # initialize mongodb client
        connection_string = (
            f'mongodb+srv://{mdb_user_token}@{mdb_hostname}'
            '/?retryWrites=true')
        self.mongo_client = pymongo.MongoClient(host=connection_string)

        # initialize mongodb database
        decimal_codec = DecimalCodec()
        type_registry = TypeRegistry([decimal_codec])
        codec_options = CodecOptions(
            tz_aware=True, 
            tzinfo=tz.tzlocal(),
            type_registry=type_registry
        )
        self.mdb = pymongo.database.Database(
            client=self.mongo_client,
            name=self.db_name,
            codec_options=codec_options)

    def create_collection(self, name):
        """Creates and returns the specified collection.

        Args:
            name: The name of the database collection to create.

        Returns:
            The MongoDB collection object.
        """
        return self.mdb.create_collection(name=name)

    def read_collection(self, name):
        """Returns the specified collection.

        Args:
            name: The name of the database collection to return.

        Returns:
            The MongoDB collection object.
        """
        return self.mdb.get_collection(name=name)

    def delete_collection(self, name):
        """Deletes the specified collection.

        Args:
            name: The name of the database collection to delete.

        Returns:
            None
        """
        self.mdb.drop_collection(name_or_collection=name)

    def __getattr__(self, name):
        return self.read_collection(name)
